unit PDBData;

interface

uses
  SysUtils, Classes, DB, RxMemDS, PDB, uImport, uExport;

type
  EventSaveReadDB = Procedure(pos, max : LongWord) of object;


  TPDBData = class(TRxMemoryData)
  private
    rc : PalmreadContext;
    ph : PDBHeader;

    FStream : TFileStream;
    FPalmFileName: string;
    FPalmDBName : string;
    FOnSavePDB: EventSaveReadDB;
    FOnReadPDB: EventSaveReadDB;
    function GetPDBCreator: string;
    function GetPDBName: string;
    function GetPDBType: string;
    procedure SetPDBCreator(const Value: string);
    procedure SetPDBName(const Value: string);
    procedure SetPDBType(const Value: string);
  protected
  public
    Constructor Create(AOwner: TComponent); override;
    destructor destroy; override;

    Procedure LoadPDB;
    Procedure SavePDB;
  published
    property PalmFileName : string read FPalmFileName write FPalmFileName;
    //property Active : boolean read GetActive write  SetActive ;
    property OnSavePDB : EventSaveReadDB read FOnSavePDB write  FOnSavePDB;
    property OnReadPDB : EventSaveReadDB read FOnReadPDB write  FOnReadPDB;
    property PDBType : string read GetPDBType write SetPDBType;
    property PDBName : string read GetPDBName write SetPDBName;
    property PDBCreator : string read GetPDBCreator write SetPDBCreator;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TPDBData]);
end;

{ TPDBData }

constructor TPDBData.Create(AOwner: TComponent);
begin
  inherited;

end;

destructor TPDBData.destroy;
begin

  inherited;
end;

{function TPDBData.GetActive: boolean;
begin
  result := inherited Active;
end;}

function TPDBData.GetPDBCreator: string;
begin
  result := ph.creator;
end;

function TPDBData.GetPDBName: string;
begin
  result := ph.name;
end;

function TPDBData.GetPDBType: string;
begin
  result := ph.ltype;
end;

procedure TPDBData.LoadPDB;
var
  i, j : integer;
  rdata : pointer;
  re1 : PDBResourceEntry ;
  re2 : PDBRecordEntry ;
  rlen : longWord;
  buf : PChar;
begin
    inherited Active := true;
    //ShowMessage(FPalmFile);
    if trim(FPalmFileName) = '' then  exit;
//        raise Exception.Create('    PALM-!' );
  try
    DisableControls;
    if not Active then exit;
    EmptyTable;
    FStream := TFileStream.Create(FPalmFileName, fmOpenRead or fmShareDenyWrite);
    //ShowMessage(FPalmFile);
    rc := palmReadHeader(FStream, ph);
    FPalmDBName := ph.name;
    //FActive := true;
    for i:=0 to ph.numRecords - 1 do
      begin
        if(rc.isResource = 1) then
          begin
            rdata := palmReadResource(FStream, rc, i, re1, rlen);
            if(rdata <> nil) then
              begin
                //stringAlong(s, re._type);
          //fprintf(fo, "\nResource %d:  Length %ld, Offset = %ld, Type \"%s\", ID = %d\n",
          //	i, rlen, rc.rsp[i].offset, s, re.id);
          //if(content)
          //	xd(fo, rdata, rlen, 1);
                FreeMem(rdata);
              end
          end
        else
          begin
            rdata := palmReadRecord(FStream, rc, i, re2, @rlen);
            buf := rdata;
            if(buf <> nil) then
              begin
                //  
                insert;
                for j:=0 to fields.Count - 1 do
                  begin
                    case fields[j].DataType of
                      ftInteger : fields[j].AsInteger := getlong(buf);
                      ftSmallint : fields[j].AsInteger := getshort(buf);
                      ftString : fields[j].AsString := GetString(buf);
                      ftBoolean : fields[j].asBoolean := GetByte(buf) = 1;
                      ftFloat : fields[j].AsFloat := GetFloat(buf);
                    end;
                  end;
                post;
                {if Assigned(FOnReadRecord) then
                  FOnReadRecord(rdata, @rlen);}
                FreeMem(rdata);
                if Assigned(FOnSavePDB) then
                  FOnReadPDB(i, ph.numRecords);
              end;
          end;
      end;
    FStream.free;
    palmDisposeContext(@rc);
    FStream := nil;
  except
    on E: Exception do
      begin
        inherited Active := false;
        raise Exception.Create(e.Message +  ':     "'+FPalmFileName+'"!' );
      end;
  end;
  EnableControls;
end;


function GetByteOfBool(val:boolean):byte;
begin
  if val then result := 1
         else result := 0;
end;

procedure TPDBData.SavePDB;
var
  i,j : integer;
  roffset : LongWord;
  StartPos : LongWord;
  reclens : Array of LongWord;
  lRecCount : LongWord;
begin
  if not Active then exit;
  if trim(FPalmFileName) = '' then
    raise Exception.Create('    PALM-!' );
  if trim(ph.name) = '' then
    raise Exception.Create('   PALM-!' );
  if trim(ph.ltype) = '' then
    raise Exception.Create('    PALM-!' );
  if trim(ph.creator) = '' then
    raise Exception.Create('    PALM-!' );

  try
    DisableControls;
    FStream := nil;
    FStream := TFileStream.Create(FPalmFileName, fmCreate);
    //rc := palmReadHeader(FStream, ph);
    //FPalmDBName := ph.name;
    lRecCount := RecordCount;
	  CreateHeader(FStream, lRecCount, ph.name, ph.ltype, ph.creator);

    // Create PDB record entry.

    roffset := kPDBHeaderSize + (lRecCount * kPDBRecordEntrySize) + 2;
    SetLength(reclens, lRecCount);
    first;
    i := 0;
    while not eof do
      begin
        outlong(FStream, roffset);			 // offset
        outlong(FStream, i);						         // attr, uniqueID:24
        reclens[i] := 0;
        for j:=0 to fields.Count - 1 do
          case fields[j].DataType of
            ftInteger  : reclens[i] := reclens[i] + 4;
            ftSmallint : reclens[i] := reclens[i] + 2;
            ftString   : reclens[i] := reclens[i] + Length(fields[j].AsString) + 1{ #0};
            ftBoolean  : reclens[i] := reclens[i] + 1;
            ftFloat    : reclens[i] := reclens[i] + 4;
          end;
        roffset := roffset + reclens[i];
        next;
        inc(i);
      end;

    outbyte(FStream, 0);
    outbyte(FStream, 0);

    // From now on we're creating the content of the database.
    first;
    for i:=0 to RecordCount - 1 do
      begin
        StartPos := FStream.Position;
        //   
        for j:=0 to fields.Count - 1 do
          case fields[j].DataType of
            ftInteger  : outlong(FStream, fields[j].Value);
            ftSmallint : outShort(FStream, fields[j].Value);
            ftString   : if fields[j].IsNull then
                           outString(FStream, '')
                         else
                           outString(FStream, fields[j].Value);
            ftBoolean  : outByte(FStream, GetByteOfBool(fields[j].Value));
            ftFloat    : outFloat(FStream, fields[j].Value);
          end;

        StartPos := FStream.Position - StartPos;
        if reclens[i] <> StartPos then
          raise Exception.Create('   !');
        if Assigned(FOnSavePDB) then
          FOnSavePDB(i, lRecCount);
         next;
      end;
  finally
    EnableControls;
    if FStream <> nil then FStream.free;
    reclens := nil;
  end;
end;

{procedure TPDBData.SetActive(const Value: boolean);
begin
  Open;
end;}

procedure TPDBData.SetPDBCreator(const Value: string);
var i : integer;
begin
  for i := 0 to 3 do
    if i < length(value) then
      ph.creator[i] := value[i]
    else
      ph.creator[i] := ' ';

end;

procedure TPDBData.SetPDBName(const Value: string);
begin
  ph.name := value;
end;

procedure TPDBData.SetPDBType(const Value: string);
var i : integer;
begin
  for i := 0 to 3 do
    if i < length(value) then
      ph.ltype[i] := value[i]
    else
      ph.ltype[i] := ' ';
end;

end.
