unit uExport;

interface
uses Classes;
type
  CalcRecLenProc = Function(index : word; fixPos:boolean = true):LongWord of object;
  ExportRecordProc = Function(Stream : TStream; index : word):LongWord of object;

  function CreateHeader(Stream : TStream; numrecords : integer; dbname, dbtype, dbcreator:string;  pdbflags:byte = 0):boolean;
  Function Export_(fn, name, type_, creator:string; records:LongWord;
    reclen : CalcRecLenProc;  exprec : ExportRecordProc; UserName:LongWord):integer;


  function outtext(Stream : TStream; text:string; len:integer):boolean;
  function outbyte(Stream : TStream; b : byte):boolean; // Two byte short
  function outshort(Stream : TStream; s : Word):boolean; // Two byte short
  function outlong(Stream : TStream; l : Longword):boolean;
  //function outbytes(Stream : TStream; buf:PChar; len:integer):boolean;
  function outbytes(Stream : TStream; Buf:PChar; len:integer):boolean;
  function outFloat(Stream : TStream; f : Single):boolean;
  function outstring(Stream : TStream; str:string):boolean;

implementation
uses PDB, SysUtils;

var
   reclens : array[0..65534] of LongWord;

function outtext(Stream : TStream; text:string; len:integer):boolean;
var
  i : integer;
  txtLen : integer;
  c : char;
begin
  txtLen := Length(text);
  c := #0;
  for i := 0 to len - 1 do
    if i < txtLen then
      Stream.WriteBuffer(text[i+1], 1)
    else
      Stream.WriteBuffer(c, 1);
  result := true;
end;

function outstring(Stream : TStream; str:string):boolean;
var
  i : integer;
  txtLen : integer;
  c : char;
begin
  txtLen := Length(str);
  c := #0;
  for i := 1 to txtLen do Stream.WriteBuffer(str[i], 1);
  Stream.WriteBuffer(c, 1); //  
  result := true;
end;

function outbyte(Stream : TStream; b : byte):boolean; // Two byte short
begin
  Stream.Write(b, 1);
  result := true;
end;


function outshort(Stream : TStream; s : Word):boolean; // Two byte short
var
  b : byte;
begin
  b := s shr 8;
  Stream.Write(b, 1);
  b := s;
  Stream.Write(b, 1);
  result := true;
end;

function outlong(Stream : TStream; l : Longword):boolean;
var
  b : byte;
begin
  b := l shr 24;
  Stream.Write(b, 1);
  b := l shr 16;
  Stream.Write(b, 1);
  b := l shr 8;
  Stream.Write(b, 1);
  b := l;
  Stream.Write(b, 1);
  result := true;
end;

function outFloat(Stream : TStream; f : Single):boolean;
var
  p : PChar;
begin
  p := @f;
  Stream.Write(p[3], 1);
  Stream.Write(p[2], 1);
  Stream.Write(p[1], 1);
  Stream.Write(p[0], 1);
  result := true;
end;

function outbytes(Stream : TStream; Buf:PChar; len:integer):boolean;
var
  i : integer;
begin
  for i:=0 to len-1 do
    Stream.Write(Buf[i], 1);
  result := true;
end;

function CreateHeader(Stream : TStream; numrecords : integer; dbname, dbtype, dbcreator:string;  pdbflags:byte = 0):boolean;
var
  today : TDateTime;
  P : Pointer;
  ltoday : LongWord;
begin
  today := now;
  P := @today;
  ltoday := LongWord(P^);

	outtext(Stream, dbname, kMaxPDBNameSize);	 // name
	outshort(Stream, pdbflags);                //	flags
	outshort(Stream, 0);                       // version
	outlong(Stream, Longword(ltoday + timeOffset));	     // creationTime
	outlong(Stream, Longword(ltoday + timeOffset));			 // modificationTime */
	outlong(Stream, 0);								         // backupTime
	outlong(Stream, 0);								         // modificationNumber */
	outlong(Stream, 0);								         // appInfoOffset */
	outlong(Stream, 0);								         // sortInfoOffset */
	outbytes(Stream, PChar(dbtype), 4);               // type
	outbytes(Stream, PChar(dbcreator), 4);					// Creator
//  outstring(Stream, 'DATA');
//  outstring(Stream, 'IDfp');

	outlong(Stream, 0);								// uniqueID
	outlong(Stream, 0);								// nextRecordID
	outshort(Stream, numrecords);			 // numRecords
  result := true;
end;

Function Export_(fn, name, type_, creator:string; records:LongWord;
  reclen : CalcRecLenProc;  exprec : ExportRecordProc; UserName:LongWord):integer;
var
  i : integer;
  Stream : TFileStream;
  roffset : LongWord;
  StartPos : LongWord;
begin

  try
    Stream := TFileStream.Create(fn, fmCreate);
  except
    result := -1;
    exit;
  end;

	CreateHeader(Stream, records, name, type_, creator);

	// Create PDB record entry.

	roffset := kPDBHeaderSize + (records * kPDBRecordEntrySize) + 2;
	for i := 0 to records - 1 do
    begin
		  outlong(Stream, roffset);			 // offset
		  outlong(Stream, i);						         // attr, uniqueID:24
      reclens[i] := reclen(i, false);
		  roffset := roffset + reclens[i];
    end;
	// Emit two pad bytes before first record. */

	outbyte(Stream, 0);
	outbyte(Stream, 0);

	// From now on we're creating the content of the database.
  for i:=0 to records - 1 do
    begin
     StartPos := Stream.Position;
     exprec(Stream, i);
     StartPos := Stream.Position - StartPos;
     if reclens[i] <> StartPos then
       raise Exception.Create('   !');
    end;
  Stream.free;
  result := 0;
end;
end.
