{
Routines for reading Palm .pdb and .prc files.

by John Walker
http://www.fourmilab.ch/
This program ("import.cpp" file) is in the public domain

Modified by Dmitry Briliuk for "Handy Finance Desktop"
http://handyware.tripod.com
bdv78@mail.ru
}
unit uImport;

interface
uses Classes, PDB, SysUtils, dialogs;
type
  ImportRecordProc = Procedure(buf:PChar; len:PLongWord);

  PpalmreadContext = ^palmreadContext;
  palmreadContext = record
        nrecords   : word;                   // Number of records
	      isResource : Integer;              // Nonzero if this is a resource file
        rsp : array of PDBResourceEntry; // Resource entry array */
	rcp : array of PDBRecordEntry;       //Record entry array */
	fileLength : Longword;                  // Total length of file */
	lengths : array of Longword;                    // Computed lengths of records */
	{ The following are not required for the functions in
           palmread.c proper but are set by them so that the
           companion byte-order-independent record cracker
           functions in palmrec.c can parse items in a
           just-returned record. }
	lastrec: PpalmreadContext;                    // Last record or resource read
	lastlength : Longword ;                  // Length of last record or resource */
	recptr : Longword ;                      // Pointer into record */
  end;

{/*  Palm file input context structure.  */
struct palmreadContext
       short nrecords;                   /* Number of records */
	int isResource;                   /* Nonzero if this is a resource file */
        struct PDBResourceEntry *rsp;     /* Resource entry array */
	struct PDBRecordEntry *rcp;       /* Record entry array */
	long fileLength;                  /* Total length of file */
	long *lengths;                    /* Computed lengths of records */
	/* The following are not required for the functions in
           palmread.c proper but are set by them so that the
           companion byte-order-independent record cracker
           functions in palmrec.c can parse items in a
           just-returned record. */
	unsigned char *lastrec;           /* Last record or resource read */
	long lastlength;                  /* Length of last record or resource */
	long recptr;                      /* Pointer into record */
}

function InByte(Stream:TStream):byte;
function inShort(Stream:TStream):Word;
function inlong(Stream:TStream):Longword;
procedure inbytes(Stream:TStream; buffer:pointer; len:integer);
function inString(Stream : TStream; len : integer) : string;


function getbyte(var buf:PChar):byte;
function getshort(var buf:PChar):word;
function getlong(var buf:PChar):LongWord;
function getfloat(var buf:PChar):Single	;
function GetString(var buf:PChar):string;

function Import(fn:string; imprec : ImportRecordProc ):integer;


function palmReadHeader(Stream : TStream;
  var ph : PDBHeader):palmreadContext;
procedure palmDisposeContext(p : PpalmreadContext);

function palmReadResource(fi:TStream; p : palmreadContext;
  resno : integer; re : PDBResourceEntry; rlen:LongWord):pointer;
function palmReadRecord(fi:TStream; p :palmreadContext; recno:LongWord;
  re : PDBRecordEntry; rlen:PLongWord):Pointer;

implementation


//FILE *fi;

///*  Byte order independent file input routines.  */

function InByte(Stream:TStream):byte;
begin
  stream.read(result,SizeOf(result));
end;

// Word  Smallint
function inShort(Stream:TStream):Word;
var
  buf2 : array[0..1] of byte;
begin
  Stream.Read(buf2,sizeOf(Buf2));
  result := (buf2[0] shl 8) or buf2[1];
end;

//long inlong()
function inlong(Stream:TStream):Longword;
var
  buf4 : array[0..3] of byte;
begin
  Stream.Read(buf4,sizeOf(Buf4));
  result := (buf4[0] shl 24) or (buf4[1] shl 16) or (buf4[2] shl 8) or buf4[3];
end;

procedure inbytes(Stream:TStream; buffer:pointer; len:integer);
begin
  Stream.ReadBuffer(buffer,len);
end;

function inString(Stream : TStream; len : integer) : string;
var
  buf : Array[0..255] of char;
  i : integer;
begin
  result := '';
  //GetMem(buf, len);
  //SetLength(buf,len);
  Stream.Read(buf, len);
  i:=0;
  while (buf[i]<>#0)and(i<len) do
    begin
      result := result + buf[i];
      inc(i);
    end;
//  SetLength(buf,0);
end;

{int getbyte(unsigned char * &p)
{
	return (unsigned char) *(p++);
}
function getbyte(var buf:PChar):byte;
begin
  result := byte(buf^);
  buf := buf + 1;
end;

{short getshort(unsigned char * &p)
{
	short s = (p[0] << 8) | p[1];
	p += 2;
	return s;
}
function getshort(var buf:PChar):word;
begin
  result := byte(buf^) shl 8;
  buf := buf + 1;
  result := result or byte(buf^);
  buf := buf + 1;
end;


{long getlong(unsigned char * &p)
{
	long l = (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
	p += 4;
	return l;
}
function getlong(var buf:PChar):LongWord;
begin
  result := byte(buf^) shl 24;
  buf := buf + 1;
  result := result or (byte(buf^) shl 16);
  buf := buf + 1;
  result := result or (byte(buf^) shl 8);
  buf := buf + 1;
  result := result or byte(buf^);
  buf := buf + 1;
end;

{float getfloat(unsigned char * &p)
{
	float f;
	*((long* )&f) = (p[0] << 24) | (p[1] << 16) | (p[2] << 8) | p[3];
	p += 4;
	return f;
}
function getfloat(var buf:PChar):Single;
var
  f : Single;
  p : pointer;
begin
  p := @f;
  LongWord(P^) := ((byte(buf[0]) shl 24) or (byte(buf[1]) shl 16) or
    (byte(buf[2]) shl 8) or byte(buf[3]));
{  result := (byte(buf[0]) shl 24) or (byte(buf[1]) shl 16) or
    (byte(buf[2]) shl 8) or byte(buf[3]);}
  result := f;
  buf := buf + 4;
end;


{char * getstring(unsigned char * &p)
{
	char * s = p;
	p += strlen(s)+1;
	return s;
}
function GetString(var buf:PChar):string;
begin
  result := '';
  while buf^<>#0 do
    begin
      result := result + buf^;
      inc(buf);
    end;

end;



{  palmReadHeader  --	Initialise to read PDB/PRC file.  The
			header is read into a buffer supplied
			by the caller and a read context is
			returned which the caller may use to
			access the contents of the file.  }
function palmReadHeader(Stream : TStream;
  var ph : PDBHeader):palmreadContext;
//struct palmreadContext *palmReadHeader(PDBHeader *ph)
var
  p : palmreadContext;
  i : integer;
  l : LongWord;

begin
{	/*	Determine the length of the file.  We need to know this
	in order to determine the length of the last record
	in the file.  */}
//	fseek(fi, 0L, SEEK_END);
  p.fileLength := Stream.Size;
//	rewind(fi);

//	/*  Read the header into the caller's PDBHeader structure.  */
  ph.name := inString(Stream, kMaxPDBNameSize);
//	inbytes(p, ph->name, kMaxPDBNameSize);
  ph.flags := inshort(Stream);
  ph.version := inshort(Stream);
  ph.creationTime := inlong(Stream);
  ph.modificationTime := inlong(Stream);
  ph.backupTime := inlong(Stream);
  ph.modificationNumber := inlong(Stream);
  ph.appInfoOffset := inlong(Stream);
  ph.sortInfoOffset := inlong(Stream);
  ph.ltype[0] := char(inByte(Stream));
  ph.ltype[1] := char(inByte(Stream));
  ph.ltype[2] := char(inByte(Stream));
  ph.ltype[3] := char(inByte(Stream));
  ph.creator[0] := char(inByte(Stream));
  ph.creator[1] := char(inByte(Stream));
  ph.creator[2] := char(inByte(Stream));
  ph.creator[3] := char(inByte(Stream));
  ph.uniqueID := inlong(Stream);
  ph.nextRecordID := inlong(Stream);
  ph.numRecords := inshort(Stream);

  p.nrecords := ph.numRecords;

  p.isResource := (ph.flags and pdbResourceFlag);
  p.rsp := nil;
  p.rcp := nil;

  SetLength(p.lengths, p.nrecords);
  //p.lengths := new long[p.nrecords];

  if p.isResource = 1 then
    begin
      SetLength(p.rsp, p.nrecords);
      //p.rsp = new PDBResourceEntry[p.nrecords];
      for i:=0 to p.nrecords -1  do
        begin
          p.rsp[i]._type := inlong(Stream);
          p.rsp[i].id := inshort(Stream);
          p.rsp[i].offset := inlong(Stream);
          if i>0 then
            p.lengths[i-1] := p.rsp[i].offset - p.rsp[i-1].offset;

        end;

      if (p.nrecords > 0) then
        p.lengths[p.nrecords-1] := p.fileLength - p.rsp[p.nrecords-1].offset
    end
  else
    begin
      SetLength(p.rcp, p.nrecords);
      for i := 0 to p.nrecords - 1 do
        begin
          p.rcp[i].offset := inlong(Stream);
          //*printf("Record %d offset = %d\n", i, p.rcp[i].offset);*/
          l := inlong(Stream);
          p.rcp[i].attr := (l shr 24) and $0FF;
          p.rcp[i].uniqueID := l and $FFFFFF;
          if (i > 0) then
            p.lengths[i-1] := p.rcp[i].offset - p.rcp[i-1].offset;
        end;
      if(p.nrecords > 0) then
        p.lengths[p.nrecords-1] := p.fileLength - p.rcp[p.nrecords-1].offset;
    end;
  p.lastrec := nil;
  p.lastlength := 0;
  p.recptr := 1;
  result := p;
end;

{  palmReadResource  --  Read a resource from a PDB/PRC file and
			  return in a newly-allocated buffer (which
                          it's the responsibility of the caller to
			  release).  }

function palmReadResource(fi:TStream; p : palmreadContext;
  resno : integer; re : PDBResourceEntry; rlen:LongWord):pointer;
//char * palmReadResource(struct palmreadContext *p, int resno, PDBResourceEntry *re, long *rlen)
var
  rec : Array of byte;
begin
  if p.isResource <> 1 then exit;
  if not ((resno >= 0) and (resno < p.nrecords)) then exit;
//  	assert(p.isResource);
//	assert((resno >= 0) && (resno < p.nrecords));

  SetLength(rec,p.lengths[resno]);

//  fseek(fi, p.rsp[resno].offset, SEEK_SET);
  fi.Position := p.rsp[resno].offset;
//  fread(rec, p.lengths[resno], 1, fi);
  fi.ReadBuffer(rec, p.lengths[resno]);
  //memcpy(re, &(p.rsp[resno]), sizeof(struct PDBResourceEntry));
  re := p.rsp[resno];
//	*rlen = p.lengths[resno];
  rlen := p.lengths[resno];
  p.lastrec := @rec;
  p.lastlength := p.lengths[resno];
  p.recptr := 0;

  result := rec;
end;

{  palmReadRecord  --	Read a resource from a PDB file and
			return in a newly-allocated buffer (which
						it's the responsibility of the caller to
			release).  }

//char * palmReadRecord(struct palmreadContext *p, int recno, PDBRecordEntry *re, long *rlen)
function palmReadRecord(fi:TStream; p :palmreadContext; recno:LongWord;
  re : PDBRecordEntry; rlen:PLongWord):Pointer;
var
  rec : pointer;
begin
  if p.isResource <> 0 then exit;
  if not ((recno >= 0) and (recno < p.nrecords)) then exit;

  //rec = new char[p.lengths[recno]];
  //SetLength(rec, p.lengths[recno]);
  GetMem(rec, p.lengths[recno]);
  fi.Position := p.rcp[recno].offset;
  fi.ReadBuffer(rec^, p.lengths[recno]);
  re := p.rcp[recno];
  rlen^ := p.lengths[recno];
  p.lastrec := rec;
  p.lastlength := p.lengths[recno];
  p.recptr := 0;


  result := rec;
end;

{  palmDisposeContext	--  Dispose of context, releasing
			    all storage.  }
procedure palmDisposeContext(p : PpalmreadContext);
begin
  try
    if p.rsp<>nil then
      SetLength(p.rsp,0);
    if p.rcp<>nil then
      SetLength(p.rcp,0);

    if p.lengths<>nil then
      SetLength(p.lengths,0);
  except
    on E: Exception do
      raise Exception.Create(e.Message +'  palmDisposeContext');
  end;
//  Dispose(p);
end;


//	Extract a 4 character string jammed into a long.

{#define stringAlong(s, l)       \
	s[0] = (char) (l >> 24);    \
	s[1] = (char) (l >> 16);    \
	s[2] = (char) (l >> 8);     \
	s[3] = (char) (l);          \
	s[4] = 0}

///*  Main program.  */

function Import(fn:string; imprec : ImportRecordProc ):integer;
//int Import(char *fn, ImportRecordProc imprec)
var
  //int i;
  i : integer;
  rc : PalmreadContext;
  ph : PDBHeader;
  dbname : string[kMaxPDBNameSize + 2];
  Stream : TFileStream;
  rdata : pointer;
  rlen : longWord;
  re1 : PDBResourceEntry ;
  re2 : PDBRecordEntry ;
  s : string[5];

begin
  try
    Stream := TFileStream.Create(fn, fmOpenRead);
  except
    result := -1;
    exit;
  end;

  rc := palmReadHeader(Stream, ph);

//	/*	Dump header.  */

  dbname := ph.name;

//	/*	Dump records or resources.  */
  for i:=0 to ph.numRecords - 1 do
    begin
      if(rc.isResource = 1) then
        begin
          rdata := palmReadResource(Stream, rc, i, re1, rlen);
          if(rdata <> nil) then
            begin
              //stringAlong(s, re._type);
				//fprintf(fo, "\nResource %d:  Length %ld, Offset = %ld, Type \"%s\", ID = %d\n",
				//	i, rlen, rc.rsp[i].offset, s, re.id);
				//if(content)
				//	xd(fo, rdata, rlen, 1);
              FreeMem(rdata);
            end
        end
      else
        begin
          rdata := palmReadRecord(Stream, rc, i, re2, @rlen);
          if(rdata <> nil) then
            begin
              imprec(rdata, @rlen);
              FreeMem(rdata);
            end;
        end;

    end;
  palmDisposeContext(@rc);
  Stream.free;
  result:=0;
end;
end.
